use std::any::Any;
use std::collections::HashMap;
use std::fmt::Debug;

// Define the Key trait with Any and Debug for type erasure and debugging
pub trait Key: Any + Debug {
    fn as_any(&self) -> &dyn Any;
}

// Implement Key for i32
impl Key for i32 {
    fn as_any(&self) -> &dyn Any {
        self
    }
}

// Implement Key for String
impl Key for String {
    fn as_any(&self) -> &dyn Any {
        self
    }
}

// Implement From<String> for i32
impl From<String> for Box<dyn Key> {
    fn from(s: String) -> Box<dyn Key> {
        if let Ok(i) = s.parse::<i32>() {
            Box::new(i)
        } else {
            Box::new(s)
        }
    }
}

// Define the Repository trait with an associated type Key
pub trait Repository {
    type Key: Key;
    fn get_entry(&self, key: &Self::Key) -> String;
}

// Define a helper trait for dynamic dispatch
pub trait DynRepository {
    fn get_entry(&self, key: &dyn Key) -> String;
}

// Define a type-erased wrapper for Repository
pub struct RepositoryWrapper<T: Repository> {
    inner: T,
}

impl<T: Repository> RepositoryWrapper<T> {
    pub fn new(inner: T) -> Self {
        Self { inner }
    }
}

impl<T: Repository> DynRepository for RepositoryWrapper<T> {
    fn get_entry(&self, key: &dyn Key) -> String {
        if let Some(key) = key.as_any().downcast_ref::<T::Key>() {
            self.inner.get_entry(key)
        } else {
            "Invalid Key Type".to_string()
        }
    }
}

// Implement the Repository trait for IntRepository
pub struct IntRepository {
    data: HashMap<i32, String>,
}

impl IntRepository {
    pub fn new() -> Self {
        let mut data = HashMap::new();
        data.insert(1, "IntEntry1".to_string());
        data.insert(2, "IntEntry2".to_string());
        Self { data }
    }
}

impl Repository for IntRepository {
    type Key = i32;

    fn get_entry(&self, key: &Self::Key) -> String {
        self.data
            .get(key)
            .unwrap_or(&"Not Found".to_string())
            .clone()
    }
}

// Implement the Repository trait for StringRepository
pub struct StringRepository {
    data: HashMap<String, String>,
}

impl StringRepository {
    pub fn new() -> Self {
        let mut data = HashMap::new();
        data.insert("one".to_string(), "StringEntry1".to_string());
        data.insert("two".to_string(), "StringEntry2".to_string());
        Self { data }
    }
}

impl Repository for StringRepository {
    type Key = String;

    fn get_entry(&self, key: &Self::Key) -> String {
        self.data
            .get(key)
            .unwrap_or(&"Not Found".to_string())
            .clone()
    }
}

fn main() {
    let int_repo = IntRepository::new();
    let string_repo = StringRepository::new();

    let repositories: Vec<Box<dyn DynRepository>> = vec![
        Box::new(RepositoryWrapper::new(int_repo)),
        Box::new(RepositoryWrapper::new(string_repo)),
    ];

    // Creating keys from strings using the From<String> implementation
    let keys: Vec<Box<dyn Key>> = vec![
        Box::new(1_i32),
        Box::new("one".to_string()),
        Box::new(99_i32),                // non-existing int key
        Box::new("unknown".to_string()), // non-existing string key
        Box::from("123".to_string()),    // i32 key from string
        Box::from("hello".to_string()),  // String key from string
    ];

    for key in keys {
        for repo in &repositories {
            let result = repo.get_entry(key.as_ref());
            println!("Result: {}", result);
        }
    }
}
